from AutoRPE.UtilsRPE.BasicStructures import Interface, Module, Routine, Function
from AutoRPE.UtilsRPE.LoadAndStore import listdir_ord
from AutoRPE.UtilsRPE.ObtainSourceFileInfo import ObtainSourcefileInfo
import AutoRPE.UtilsRPE.PointerDealer as PointerDealer
from AutoRPE.UtilsRPE.DerivedTypeDealer import DerivedType
from AutoRPE.UtilsRPE.Vault import Vault

from __future__ import print_function
import pickle
from os.path import isfile, join

def parse_sources(path_to_model_sources, save_vault=False):
    # Find the nemo source-files in sources
    # (Using the sources that already modified replacing real declarations by type(rpe_var) )
    sources = [f for f in listdir_ord(path_to_model_sources) if
               isfile(join(path_to_model_sources, f)) and f.count(".f90")]
    # Initialize two empty dictionaries that will contain all the routines and all the functions
    object_oriented_dictionary = {"procedures": {}, "interfaces": {}, "variables": {}, "modules": {},
                                  "derivedtypes": {}}

    # Initialize an empty dictionary that will contain all the source-files.
    sourcefiles = {}
    vault = Vault()

    print("\n\nParsing sourcecode...\n")

    # Going over all the source-files loading the files,
    # parsing the routines and storing both things in dictionaries.

    for index, filename in enumerate(sources):
        module_name = filename.replace(".f90", "")
        sourcefiles[module_name] = ObtainSourcefileInfo(join(path_to_model_sources, filename))
        module = Module(module_name)
        f = sourcefiles[module_name]
        f.module = module
        f.find_routines()
        f.find_functions()
        f.find_module_variables()
        for procedure in f.objects:
            object_oriented_dictionary["procedures"][procedure.name] = procedure
            module.add_procedure(procedure)
            if procedure.name == "main":
                module.variables = procedure.variables
                for var in procedure.variables:
                    vault.variables.append(var)
        vault.add_module(module)
        for procedure in f.objects:
            vault.procedures.append(procedure)
            if isinstance(procedure, Routine):
                vault.routines.append(procedure)
                for var in procedure.variables:
                    vault.variables.append(var)
            if isinstance(procedure, Function):
                vault.functions.append(procedure)
                for var in procedure.variables:
                    vault.variables.append(var)


    # Going over all the source-files parsing the interfaces and storing them in dictionary.
    for index, filename in enumerate(sources):
        module_name = filename.replace(".f90", "")
        f = sourcefiles[module_name]
        f.find_interfaces(object_oriented_dictionary["procedures"])
        # Getting and saving modules used
        f.module = vault[module_name]
        f.find_used_modules(vault)
        for interface in f.objects:
            if isinstance(interface, Interface):
                object_oriented_dictionary["interfaces"][interface.name] = interface
                vault[module_name].procedures.append(interface)
                vault.interfaces.append(interface)
                vault.procedures.append(interface)

    # After having all the used parsed, propagate the inheritance, and add public var to available vars
    vault.expand_use_statement()

    # Going over all the source-files parsing the derived types and storing them in dictionary.
    for index, filename in enumerate(sources):
        module_name = filename.replace(".f90", "")
        f = sourcefiles[module_name]
        f.find_types()
        for derived_type in f.objects:
            if isinstance(derived_type, DerivedType):
                object_oriented_dictionary["derivedtypes"][derived_type.name] = derived_type
                vault[module_name].procedures.append(derived_type)
                vault.derived_types.append(derived_type)
                vault.procedures.append(derived_type)
                for var in derived_type.variables:
                    vault.variables.append(var)

    # Fill a dictionary to speed up search of variables
    vault.process_variables()
    # Now that vault.derived_types is full add extension to derived types if any
    vault.expand_derived_types()

    # Passing to pointed objects the actual function/subroutine they points to
    PointerDealer.process_pointers(vault)

    # # Find implicit function declarations
    # for filename in sourcefiles:
    #     module_name = filename.replace(".f90", "")
    #     f = sourcefiles[module_name]
    #     f.find_implicit_functions(vault=vault)
    #     for procedure in f.objects:
    #         if procedure not in vault[module_name].procedures:
    #             vault[module_name].procedures.append(procedure)
    #             vault.procedures.append(procedure)

    print("number of functions " + str(len(vault.functions)) + "\n" +
          "number or routines " + str(len(vault.routines)) + "\n" +
          "number or interfaces " + str(len(vault.interfaces)) + "\n" +
          "number or main " + str(len(vault.modules)) + "\n" +
          "number or procedures " + str(len(vault.procedures)) + "\n" +
          "number of variables " + str(len(vault.variables)) + "\n" +
          "number or derived_types " + str(len(vault.derived_types)) + "\n")

    if save_vault:
        # This workaround is needed to allow pickle to write the file, don't know yet about its risks
        import sys
        sys.setrecursionlimit(2000)
        pickle.dump(vault, open("vault.pkl", "wb"))
    return vault


def preprocess_sources(path_to_input_sources, path_to_output_sources, blacklist=[]):
    from AutoRPE.UtilsRPE.ImplementRPE import ImplementRPE
    # Read all files sources excluding hidden files ( like vim swp files)
    all_sources = [f for f in listdir_ord(path_to_input_sources) if
                   isfile(join(path_to_input_sources, f)) and f.count(".f90") and f[0][0] != "."]
    total = len(all_sources)
    if len(all_sources):
        # Load and format files
        for index, filename in enumerate(all_sources):
            f = ImplementRPE(join(path_to_input_sources, filename), None)
            # Remove comments, empty lines and old style declarations
            if filename in blacklist:
                print("\rFormatting file %i/%i  %20s" % (index + 1, total, filename), end="")
                f.pre_process()
            # Additionally reals are replaced with rpe_var type
            else:
                print("\rFormatting file and replacing reals %i/%i  %20s" % (index + 1, total, filename), end="")
                f.pre_process(replace_reals=True)
            f.save_modified_file(join(path_to_output_sources, filename))
    else:
        # Sources is an empty folder
        raise ValueError("No files found in %s folder", path_to_input_sources)


def postprocess_sources(path_to_input_sources, path_to_output_sources, vault):
    from AutoRPE.UtilsRPE.ImplementRPE import ImplementRPE

    # List files in source sexcluding hidden files ( like vim swp files)
    all_sources = [f for f in listdir_ord(path_to_input_sources) if
                   isfile(join(path_to_input_sources, f)) and f.count(".f90")and f[0][0] != "."]
    total = len(all_sources)
    for index, filename in enumerate(all_sources):
        f = ImplementRPE(join(path_to_input_sources, filename), vault)
        print("\rFixing issues %i/%i  %20s" % (index + 1, total, filename), end="")
        f.post_process()
        f.save_modified_file(join(path_to_output_sources, filename))


